import ast
import datetime
import functools
from typing import Iterable, Tuple, List, Union, Dict, Callable, Optional

import arrow
import graphene
import django_filters
import stringcase
from arrow import Arrow
from graphene import Scalar
from graphql.language import ast
from graphene_django import DjangoObjectType
from graphene_django_extras import LimitOffsetGraphqlPagination, DjangoInputObjectType, DjangoListObjectType

from django_koldar_utils.django import django_helpers
from rest_framework import serializers

from django_koldar_utils.graphql.scalars.ArrowDateScalar import ArrowDateScalar
from django_koldar_utils.graphql.scalars.ArrowDateTimeScalar import ArrowDateTimeScalar
from django_koldar_utils.graphql.scalars.ArrowDurationScalar import ArrowDurationScalar


def create_graphene_tuple_type(name: str, it: Iterable[Tuple[str, type]]) -> type:
    """
    Programmatically create a type in graphene repersenting a tuple of elements.

    :param name: name of the type
    :param it: an iteralbe of pairs, where the frist item is the field name repersenting the i-th element
        while the second item is the graphene.FIeld type of said graphene class field
    :return: type rerpesenting the tuple
    """
    properties = dict()
    for key, type in it:
        properties[key] = type
    result = type(
        name,
        (graphene.ObjectType, ),
        properties
    )
    return result


def create_graphene_pair_type(name: str, item0_name: str, item0_type: type, item1_name: str, item1_type: type) -> type:
    """
    Programmatically create a type in graphene repersenting a pair of elements.

    :param name: name of the type
    :param item0_name: field name of the first item
    :param item0_type: field graphene.FIeld type of the first item
    :param item1_name: field name of the second item
    :param item1_type: field graphene.FIeld type of the second item
    :return: type rerpesenting the tuple
    """
    return create_graphene_tuple_type(name, [(item0_name, item0_type), (item1_name, item1_type)])


def create_graphql_class(cls, fields=None, specify_fields: Dict[str, Tuple[type, Optional[Callable[[any, any], any]]]]=None) -> type:
    """
    Create a graphQl type starting from a Django model

    :param cls: django type of the model whose graphql type we want to generate
    :param fields: field that we wna tto include in the graphene class type
    :param specify_fields: a dictionary of django model fields which you want to personally customize.
        Each dictionary key is a django model field name. Each value is a pair. If present, "fields" is ignored
         - first, mandatory, is the graphene type that you want to use for the field
         - second, (optionally set to None) is a callable representing the resolver. If left missing we will just call
            the model field
    """

    def default_resolver(model_instance, info, field_name: str = None) -> any:
        return getattr(model_instance, field_name)

    if fields is None:
        fields = "__all__"
    if specify_fields is None:
        specify_fields = dict()

    meta_properties = {
        "model": cls,
        "description": cls.__doc__,
    }
    if len(specify_fields) > 0:
        meta_properties["exclude"] = list(specify_fields.keys())
    else:
        meta_properties["fields"] = fields
    graphql_type_meta = type(
        "Meta",
        (object, ),
        meta_properties
    )

    class_name = cls.__name__
    properties = {
        "Meta": graphql_type_meta,
    }
    # attach graphql type additional fields
    for field_name, value in specify_fields.items():
        if isinstance(value, tuple):
            graphene_type, resolver_function = value
        else:
            graphene_type = value
            resolver_function = None

        properties[field_name] = graphene_type
    for field_name, value in specify_fields.items():
        if isinstance(value, tuple):
            graphene_type, resolver_function = value
        else:
            graphene_type = value
            resolver_function = None

        if resolver_function is None:
            resolver_function = functools.partial(default_resolver, field_name=field_name)
        properties[f"resolve_{field_name}"] = resolver_function

    graphql_type = type(
        f"{class_name}GraphQLType",
        (DjangoObjectType, ),
        properties
    )

    return graphql_type


def create_graphql_primitive_input(django_type: type, graphene_type: type, exclude_fields: List[str] = None) -> type:
    """
    Create an input class from a django model specifying only primitive types.
    All such types are optional (not required)
    """

    # class PersonInput(graphene.InputObjectType):
    #     name = graphene.String(required=True)
    #     age = graphene.Int(required=True)

    if exclude_fields is None:
        exclude_fields = []

    class_name = django_type.__name__
    description = f"""The graphql input tyep associated to the type {class_name}. See {class_name} for further information"""

    primitive_fields = {}
    for field in django_helpers.get_primitive_fields(django_type):
        if field.attname in exclude_fields:
            continue
        t = graphene_type._meta.fields[field.attname]
        if isinstance(t, graphene.Field):
            graphene_field_type = t.type.of_type
        else:
            graphene_field_type = t.type

        if graphene_field_type._meta.name == "String":
            v = graphene.String(required=False, default_value=t.default_value, description=t.description, **t.args)
        elif graphene_field_type._meta.name == "Int":
            v = graphene.Int(required=False, default_value=t.default_value, description=t.description, **t.args)
        elif graphene_field_type._meta.name == "Boolean":
            v = graphene.Boolean(required=False, default_value=t.default_value, description=t.description, **t.args)
        elif graphene_field_type._meta.name == "ID":
            v = graphene.ID(required=False, default_value=t.default_value, description=t.description, **t.args)
        elif graphene_field_type._meta.name == "DateTime":
            v = graphene.DateTime(required=False, default_value=t.default_value, description=t.description, **t.args)
        elif graphene_field_type._meta.name == "Date":
            v = graphene.Date(required=False, default_value=t.default_value, description=t.description, **t.args)
        elif graphene_field_type._meta.name == "ArrowDateTimeScalar":
            v = ArrowDateTimeScalar(required=False, default_value=t.default_value, description=t.description, **t.args)
        elif graphene_field_type._meta.name == "ArrowDateScalar":
            v = ArrowDateScalar(required=False, default_value=t.default_value, description=t.description, **t.args)
        elif graphene_field_type._meta.name == "ArrowDurationScalar":
            v = ArrowDurationScalar(required=False, default_value=t.default_value, description=t.description, **t.args)
        elif graphene_field_type._meta.name == "Base64":
            v = graphene.Base64(required=False, default_value=t.default_value, description=t.description, **t.args)
        elif graphene_field_type._meta.name == "Float":
            v = graphene.Float(required=False, default_value=t.default_value, description=t.description, **t.args)
        else:
            raise ValueError(f"cannot handle type {t} (name = {graphene_field_type._meta.name})!")

        primitive_fields[field.attname] = v

    input_graphql_type = type(
        f"{stringcase.pascalcase(class_name)}PrimitiveGraphQLInput",
        (graphene.InputObjectType, ),
        {
            "model": django_type,
            "description": description,
            "__doc__": description,
            **primitive_fields
        }
    )

    return input_graphql_type


def create_graphql_input(cls) -> type:
    """
    See dujango extras
    """

    graphql_type_meta = type(
        "Meta",
        (object, ),
        {
            "model": cls,
            "description": f"""
                Input type of class {cls.__name__}.
            """
        }
    )

    class_name = cls.__name__
    graphql_type = type(
        f"{class_name}GraphQLInput",
        (DjangoInputObjectType, ),
        {
            "Meta": graphql_type_meta
        }
    )

    return graphql_type


def create_graphql_list_type(cls) -> type:
    """
    A graphql type representing a list of a given class.
    This is used to generate list of DjancoObjectType
    See https://github.com/eamigo86/graphene-django-extras
    """
    graphql_type_meta = type(
        "Meta",
        (object, ),
        {
            "model": cls,
            "description": f"""GraphQL type representing a list of {cls.__name__}.""",
            "pagination": LimitOffsetGraphqlPagination(default_limit=25)
        }
    )

    class_name = cls.__name__
    graphql_type = type(
        f"{class_name}GraphQLListType",
        (DjangoListObjectType, ),
        {
            "Meta": graphql_type_meta
        }
    )
    return graphql_type



def create_serializer(cls) -> type:
    """
    A serializer allowing to easily create mutations
    See https://github.com/eamigo86/graphene-django-extras
    """
    graphql_type_meta = type(
        "Meta",
        (object, ),
        {
            "model": cls,
        }
    )

    class_name = cls.__name__
    graphql_type = type(
        f"{class_name}Serializer",
        (serializers.ModelSerializer, ),
        {
            "Meta": graphql_type_meta
        }
    )
    return graphql_type